package cn.dankal.share.concurrentutil;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * token Bucket 简单模拟
 * <p>
 * 基于 {@link Semaphore} 简单实现
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class TokenBucket {

    /**
     * <code>Semaphore</code>
     */
    private final Semaphore semaphore;

    /**
     * <code>CountDownLatch</code>
     */
    private final CountDownLatch countDownLatch;

    /**
     * 令牌统计
     */
    private final AtomicInteger tokenCounter = new AtomicInteger(0);

    /**
     * 成功线程数统计
     */
    private final AtomicInteger successCounter = new AtomicInteger(0);

    /**
     * 失败线程数统计
     */
    private final AtomicInteger fairCounter = new AtomicInteger(0);

    public TokenBucket(Semaphore semaphore, CountDownLatch countDownLatch1) {
        this.semaphore = semaphore;
        this.countDownLatch = countDownLatch1;
    }

    /**
     * 获取令牌
     */
    public void getToken() {
        try {
            final long getTokenTimeout = 5;
            if (semaphore.tryAcquire(getTokenTimeout, TimeUnit.MILLISECONDS)) {
                // 成功获取到信号量: 获取令牌成功
                // 保证 tryAcquire 超时
                int tokenNum = tokenCounter.incrementAndGet();
                TimeUnit.SECONDS.sleep(1);
                System.out.printf("%s 成功获取到 %d 号令牌 %n", Thread.currentThread().getName(), tokenNum);
                successCounter.incrementAndGet();
            } else {
                // 获取令牌失败
                System.out.printf("%s 未获取到令牌 %n", Thread.currentThread().getName());
                fairCounter.incrementAndGet();
            }
        } catch (InterruptedException e) {
            // ignored
        } finally {
            // 成功/失败统计计数扣减
            this.countDownLatch.countDown();
        }

    }

    /**
     * 成功获取令桶统计
     *
     * @return 成功获取令牌统计
     */
    public int successCount() {
        return successCounter.get();
    }

    /**
     * 获取令牌失败统计
     *
     * @return 未成功获取令牌统计
     */
    public int fairCount() {
        return fairCounter.get();
    }
}
