package cn.dankal.share.util.completablefuture;

import cn.dankal.share.common.ThreadExecutorFactory;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.core.util.StrUtil;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Test;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class CompletableFutureMethodTest {

    /**
     * ThreadPoolExecutor
     */
    private final static ThreadPoolExecutor THREAD_POOL_EXECUTOR = ThreadExecutorFactory.getThreadPoolExecutor("completableFuture-test");


    /**
     * 测试 CompletableFuture: thenApply(Function<? super T,? extends U> fn)
     */
    @Test
    public void testThenApply() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        CompletableFuture<Integer> completableFuture = CompletableFuture
                .supplyAsync(() -> {
                    try {
                        TimeUnit.SECONDS.sleep(1);
                    } catch (InterruptedException e) {
                        // ignored
                    }
                    int randomInt = RandomUtil.randomInt(1, 100);
                    System.out.printf("%s 成功生产了 %d%n", Thread.currentThread().getName(), randomInt);
                    return randomInt;
                }, THREAD_POOL_EXECUTOR)
                .thenApply(result -> {
                    System.out.printf("%s 成功获取结果,耗时 %d 秒%n", Thread.currentThread().getName(), (System.currentTimeMillis() - startMs) / 1000);
                    System.out.printf("%s 线程即将消费 %d%n", Thread.currentThread().getName(), result);
                    return result + 100;
                });
        Integer result = completableFuture.get();
        System.out.println("result = " + result);
    }

    /**
     * 测试 CompletableFuture: thenApplyAsync(Function<? super T,? extends U> fn, Executor executor)
     */
    @Test
    public void testThenApplyAsync() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        CompletableFuture<Integer> completableFuture = CompletableFuture
                .supplyAsync(() -> {
                    try {
                        TimeUnit.SECONDS.sleep(1);
                    } catch (InterruptedException e) {
                        // ignored
                    }
                    int randomInt = RandomUtil.randomInt(1, 100);
                    System.out.printf("%s 成功生产了 %d%n", Thread.currentThread().getName(), randomInt);
                    return randomInt;
                }, THREAD_POOL_EXECUTOR)
                .thenApplyAsync(result -> {
                    System.out.printf("%s 成功获取结果,耗时 %d 秒%n", Thread.currentThread().getName(), (System.currentTimeMillis() - startMs) / 1000);
                    System.out.printf("%s 线程即将消费 %d%n", Thread.currentThread().getName(), result);
                    return result + 100;
                }, THREAD_POOL_EXECUTOR);
        Integer result = completableFuture.get();
        System.out.println("result = " + result);
    }


    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *       thenCombineAsync(
     *         CompletionStage<? extends U> other,
     *         BiFunction<? super T,? super U,? extends V> fn, Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testThenCombineAsync() throws ExecutionException, InterruptedException {
        CompletableFuture<Integer> integerCompletableFuture = CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 100);
            final long sleepTime = 5;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒 %n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR);

        final long startMs = System.currentTimeMillis();
        Integer result = CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 10);
            final long sleepTime = 2;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR)
                .thenCombineAsync(integerCompletableFuture, Integer::sum)
                .get();
        System.out.printf("获取最终结果耗时 %d 秒%n", (System.currentTimeMillis() - startMs) / 1000);
        System.out.println("result = " + result);
    }

    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *      thenComposeAsync(
     *         Function<? super T, ? extends CompletionStage<U>> fn,
     *         Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testThenComposeAsync() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        Integer result = CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 10);
            final long sleepTime = 5;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR)
                .thenComposeAsync(integer ->
                        CompletableFuture.supplyAsync(() -> {
                            System.out.printf("%s 获取到值 %d%n", Thread.currentThread().getName(), integer);
                            int randomInt = RandomUtil.randomInt(1, 100);
                            final long sleepTime = 2;
                            try {
                                TimeUnit.SECONDS.sleep(sleepTime);
                            } catch (InterruptedException e) {
                                //ignored
                            }
                            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
                            return randomInt + integer;
                        }, THREAD_POOL_EXECUTOR))
                .get();
        System.out.printf("获取最终结果耗时 %d 秒%n", (System.currentTimeMillis() - startMs) / 1000);
        System.out.println("result = " + result);
    }


    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *      thenAcceptBothAsync(
     *         CompletionStage<? extends U> other,
     *         BiConsumer<? super T, ? super U> action, Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testThenAcceptBothAsync() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 10);
            final long sleepTime = 5;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR)
                .thenAcceptBothAsync(
                        CompletableFuture.supplyAsync(() -> {
                            int randomInt = RandomUtil.randomInt(1, 100);
                            final long sleepTime = 2;
                            try {
                                TimeUnit.SECONDS.sleep(sleepTime);
                            } catch (InterruptedException e) {
                                //ignored
                            }
                            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
                            return randomInt;
                        }), (result1, result2) -> System.out.printf("两个 CompletableFuture 和: %d%n", result1 + result2), THREAD_POOL_EXECUTOR)
                .get();
        System.out.printf("获取最终结果耗时 %d 秒%n", (System.currentTimeMillis() - startMs) / 1000);
    }

    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *      runAfterBothAsync(
     *                         CompletionStage<?> other,
     *                         Runnable action,
     *                         Executor executor
     *                         )
     *     }
     * </pre>
     */
    @Test
    public void testRunAfterBothAsync() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 10);
            final long sleepTime = 5;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR)
                .runAfterBothAsync(
                        CompletableFuture.supplyAsync(() -> {
                            int randomInt = RandomUtil.randomInt(1, 100);
                            final long sleepTime = 2;
                            try {
                                TimeUnit.SECONDS.sleep(sleepTime);
                            } catch (InterruptedException e) {
                                //ignored
                            }
                            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
                            return randomInt;
                        }, THREAD_POOL_EXECUTOR), () -> System.out.printf("%s :两个 CompletableFuture运行耗时 %d 秒%n", Thread.currentThread().getName(), (System.currentTimeMillis() - startMs) / 1000)
                        , THREAD_POOL_EXECUTOR)
                .get();
    }

    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *      applyToEitherAsync(
     *         CompletionStage<? extends T> other, Function<? super T, U> fn,
     *         Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testApplyToEitherAsync() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        Integer result = CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 10);
            final long sleepTime = 6;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR)
                .applyToEitherAsync(
                        CompletableFuture.supplyAsync(() -> {
                            int randomInt = RandomUtil.randomInt(1, 100);
                            final long sleepTime = 2;
                            try {
                                TimeUnit.SECONDS.sleep(sleepTime);
                            } catch (InterruptedException e) {
                                //ignored
                            }
                            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
                            return randomInt;
                        }, THREAD_POOL_EXECUTOR), (i) -> {
                            System.out.printf("%s :获取到结果: %d 等待耗时 %d 秒%n", Thread.currentThread().getName(), i, (System.currentTimeMillis() - startMs) / 1000);
                            final long sleepTime = 3;
                            try {
                                TimeUnit.SECONDS.sleep(sleepTime);
                            } catch (InterruptedException e) {
                                // ignored
                            }
                            System.out.printf("%s 处理结果耗时 %d 秒%n", Thread.currentThread().getName(), sleepTime);
                            return i + 100;
                        }
                        , THREAD_POOL_EXECUTOR)
                .get();
        System.out.printf("获取到最终结果: %d, 耗时 %d 秒 %n", result, (System.currentTimeMillis() - startMs) / 1000);
    }

    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *      acceptEitherAsync(
     *         CompletionStage<? extends T> other, Consumer<? super T> action,
     *         Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testAcceptEitherAsync() throws ExecutionException, InterruptedException {
        final long startMs = System.currentTimeMillis();
        CompletableFuture.supplyAsync(() -> {
            int randomInt = RandomUtil.randomInt(1, 10);
            final long sleepTime = 5;
            try {
                TimeUnit.SECONDS.sleep(sleepTime);
            } catch (InterruptedException e) {
                //ignored
            }
            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
            return randomInt;
        }, THREAD_POOL_EXECUTOR)
                .acceptEitherAsync(
                        CompletableFuture.supplyAsync(() -> {
                            int randomInt = RandomUtil.randomInt(1, 100);
                            final long sleepTime = 2;
                            try {
                                TimeUnit.SECONDS.sleep(sleepTime);
                            } catch (InterruptedException e) {
                                //ignored
                            }
                            System.out.printf("%s 生产了 %d 耗时 %d 秒%n", Thread.currentThread().getName(), randomInt, sleepTime);
                            return randomInt;
                        }, THREAD_POOL_EXECUTOR), (result) -> System.out.printf("%s :获取到结果: %d 等待耗时 %d 秒%n", Thread.currentThread().getName(), result, (System.currentTimeMillis() - startMs) / 1000)
                        , THREAD_POOL_EXECUTOR)
                .get();
    }


    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *       exceptionally(
     *         Function<Throwable, ? extends T> fn)
     *     }
     * </pre>
     */
    @Test
    public void testExceptionally() throws ExecutionException, InterruptedException {
        String result = CompletableFuture.
                supplyAsync(() -> "成功响应",
                        THREAD_POOL_EXECUTOR)
                .exceptionally(Throwable::getMessage)
                .get();
        System.out.printf("成功响应结果: %s%n", result);
        result = CompletableFuture.
                supplyAsync(() -> {
                            if (true) {
                                throw new RuntimeException("测试 exceptional 异常补偿");
                            }
                            return "成功响应";
                        },
                        THREAD_POOL_EXECUTOR)
                .exceptionally(Throwable::getMessage)
                .get();
        System.out.printf("失败响应结果: %s%n", result);

    }


    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *       whenCompleteAsync(
     *         BiConsumer<? super T, ? super Throwable> action, Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testWhenCompeteAsync() throws ExecutionException, InterruptedException {
        String result = CompletableFuture.supplyAsync(() -> "dankal", THREAD_POOL_EXECUTOR)
                .whenCompleteAsync((str, exception) -> {
                    if (StrUtil.isNotBlank(str)) {
                        System.out.println("响应结果: " + str);
                    }
                    if (null != exception) {
                        System.out.println("发生异常: " + exception.getMessage());
                    }
                }, THREAD_POOL_EXECUTOR)
                .get();
        System.out.println("result = " + result);
        result = CompletableFuture.supplyAsync(() -> {
            if (true) {
                throw new RuntimeException("just throw a exception");
            }
            return "dankal";
        }, THREAD_POOL_EXECUTOR)
                .whenCompleteAsync((str, exception) -> {
                    if (StrUtil.isNotBlank(str)) {
                        System.out.println("响应结果: " + str);
                    }
                    if (null != exception) {
                        System.out.println("发生异常: " + exception.getMessage());
                    }
                }, THREAD_POOL_EXECUTOR)
                .get();
        System.out.println("result = " + result);
    }

    /**
     * 测试 CompletableFuture:
     * <pre>
     *     {@code
     *      handleAsync(
     *         BiFunction<? super T, Throwable, ? extends U> fn, Executor executor)
     *     }
     * </pre>
     */
    @Test
    public void testHandlerAsync() throws ExecutionException, InterruptedException {
        String resultStr = CompletableFuture.supplyAsync(() -> "success", THREAD_POOL_EXECUTOR)
                .handleAsync((res, exception) -> {
                    if (StrUtil.isNotBlank(res)) {
                        return res;
                    }
                    return exception.getMessage();
                }, THREAD_POOL_EXECUTOR)
                .get();
        System.out.println("resultStr = " + resultStr);
        resultStr = CompletableFuture.supplyAsync(() -> {
            if (true) {
                throw new RuntimeException("a exception happened");
            }
            return "success";
        }, THREAD_POOL_EXECUTOR)
                .handleAsync((res, exception) -> {
                    if (StrUtil.isNotBlank(res)) {
                        return res;
                    }
                    return exception.getMessage();
                }, THREAD_POOL_EXECUTOR)
                .get();
        System.out.println("resultStr = " + resultStr);

    }

    /**
     * 关闭 线程池资源
     */
    @AfterAll
    public static void after() {
        THREAD_POOL_EXECUTOR.shutdown();
    }
}
